{% if page.meta.source %}
<div class="source-links">
{% for filename in page.meta.source %}
    <span class="label label-primary">{{ filename }}</span>
{% endfor %}
</div>
{% endif %}

{{ page.content }}
