body {
    padding-top: 50px;
}

/* Replacement for `body { background-attachment: fixed; }`, which has
   performance issues when scrolling on large displays. See #1394. */
body::before {
    content: ' ';
    position: fixed;
    width: 100%;
    height: 100%;
    top: 0;
    left: 0;
    background-color: #f8f8f8;
    background: url(../images/mkdocs-grid.png) repeat-x;
    will-change: transform;
    z-index: -1;
}

body > .container {
    min-height: 400px;
}

ul.nav .main {
    font-weight: bold;
}

.source-links {
    float: right;
}

.col-md-9 img {
    max-width: 100%;
    display: inline-block;
    padding: 4px;
    line-height: 1.428571429;
    background-color: #fff;
    /* border: 1px solid #ddd; */
    border-radius: 4px;
    /* margin: 20px auto 30px auto; */
}

/*
 * The code below adds some padding to the top of the current anchor target so
 * that, when navigating to it, the header isn't hidden by the navbar at the
 * top.
 */
:target::before {
    content: "";
    display: block;
    margin-top: -75px;
    height: 75px;
    pointer-events: none;
}

h1 {
    color: #444;
    font-weight: 400;
    font-size: 42px;
}

h2, h3, h4, h5, h6 {
    color: #444;
    font-weight: 300;
}

hr {
    border-top: 1px solid #aaa;
}

pre, .rst-content tt {
    max-width: 100%;
    background: #fff;
    color: #333;
    overflow-x: auto;
}

code.code-large, .rst-content tt.code-large {
    font-size: 90%;
}

code {
    padding: 2px 5px;
    background: #fff;
    border: solid 1px #e1e4e5;
    color: #333;
    white-space: pre-wrap;
    word-wrap: break-word;
}

pre code {
    display: block;
    background: transparent;
    border: none;
    white-space: pre;
    word-wrap: normal;
    font-family: SFMono-Regular, Menlo, Monaco, Consolas, "Liberation Mono", "Courier New", monospace;
    font-size: 12px;
}

kbd {
    padding: 2px 4px;
    font-size: 90%;
    color: #fff;
    background-color: #333;
    border-radius: 3px;
    -webkit-box-shadow: inset 0 -1px 0 rgba(0,0,0,.25);
    box-shadow: inset 0 -1px 0 rgba(0,0,0,.25);
}

a code {
    color: #2FA4E7;
}

a:hover code, a:focus code {
    color: #157AB5;
}

footer {
    margin-top: 30px;
    margin-bottom: 10px;
    text-align: center;
    font-weight: 200;
}

.modal-dialog {
    margin-top: 50px;
}

/*
 * Side navigation
 *
 * Scrollspy and affixed enhanced navigation to highlight sections and secondary
 * sections of docs content.
 */

.bs-sidebar.affix { /* csslint allow: adjoining-classes */
    /* csslint ignore:start */
    position: -webkit-sticky;
    position: sticky;
    /* csslint ignore:end */
    top: 80px;
}

.bs-sidebar.card { /* csslint allow: adjoining-classes */
    padding: 0;
    max-height: 90%;
    overflow-y: auto;
}

/* Toggle (vertically flip) sidebar collapse icon */
.bs-sidebar .navbar-toggler span {
    -moz-transform: scale(1, -1);
    -webkit-transform: scale(1, -1);
    -o-transform: scale(1, -1);
    -ms-transform: scale(1, -1);
    transform: scale(1, -1);
}

.bs-sidebar .navbar-toggler.collapsed span { /* csslint allow: adjoining-classes */
    -moz-transform: scale(1, 1);
    -webkit-transform: scale(1, 1);
    -o-transform: scale(1, 1);
    -ms-transform: scale(1, 1);
    transform: scale(1, 1);
}

/* First level of nav */
.bs-sidenav {
    padding-top:    10px;
    padding-bottom: 10px;
    border-radius: 5px;
    width: 100%;
}

/* All levels of nav */
.bs-sidebar .nav > li > a {
    display: block;
    padding: 3px 15px;
    z-index: 1;
}
.bs-sidebar .nav > li > a:hover,
.bs-sidebar .nav > li > a:focus {
    text-decoration: none;
    border-right: 2px solid;
}
.bs-sidebar .nav > li > a.active,
.bs-sidebar .nav > li > a.active:hover,
.bs-sidebar .nav > li > a.active:focus {
    font-weight: bold;
    background-color: transparent;
    border-right: 2px solid;
}

/* Nav: second level (shown on .active) */
.bs-sidebar .nav .nav {
    display: none; /* Hide by default, but at >768px, show it */
    margin-bottom: 8px;
}
.bs-sidebar .nav .nav > li > a {
    padding-top:    3px;
    padding-bottom: 3px;
    padding-left: 30px;
    font-size: 90%;
}

.headerlink {
    font-family: FontAwesome;
    font-size: 14px;
    display: none;
    padding-left: .5em;
}

h1:hover .headerlink, h2:hover .headerlink, h3:hover .headerlink, h4:hover .headerlink, h5:hover .headerlink, h6:hover .headerlink{
    display:inline-block;
}



.admonition {
    padding: 15px;
    margin-bottom: 10px;
    border: 1px solid transparent;
    border-radius: 4px;
    text-align: left;
}

.admonition.caution { /* csslint allow: adjoining-classes */
    color: #000000;
    background-color: #ffcc44;
    border: 1px dashed #ff8800;
    padding: .5em;
}

.admonition.convention { /* csslint allow: adjoining-classes */
    color: #000000;
    background-color: #ccffcc;
    border: 1px dashed #88ff88;
    padding: .5em;
}

.admonition.missing { /* csslint allow: adjoining-classes */
    color: #000000;
    background-color: #ffc0a0;
    border: 1px dashed #440000;
    padding: .5em;
}

.admonition.note { /* csslint allow: adjoining-classes */
    color: #000000;
    background-color: #ffffdd;
    border: 1px dashed #ffcc44;
    padding: .5em;
}

.admonition.note2 { /* csslint allow: adjoining-classes */
    color: #3a87ad;
    background-color: #d9edf7;
    border-color: #bce8f1;
}

.admonition.conference { /* csslint allow: adjoining-classes */
    color: #000000;
    background-color: rgb(171,221,139);
    border-color: #338033;
}

.admonition.convention {
    color: #000000;
    background-color: #ccffcc; 
    padding: .5em;
    border: 1px dashed #88ff88;
}

.admonition.incomplete {
    color: #000000;
    background-color: #ffc0a0; 
    padding: .5em; 
    border: 1px dashed #440000;
}

.admonition.danger { /* csslint allow: adjoining-classes */
    color: #b94a48;
    background-color: #f2dede;
    border-color: #eed3d7;
}

.admonition-title {
    font-weight: bold;
    text-align: left;
}


.dropdown-submenu {
    position: relative;
}

.dropdown-submenu>.dropdown-menu {
    top: 0;
    left: 100%;
    margin-top: -6px;
    margin-left: -1px;
    -webkit-border-radius: 0 6px 6px 6px;
    -moz-border-radius: 0 6px 6px;
    border-radius: 0 6px 6px 6px;
}

.dropdown-submenu:hover>.dropdown-menu {
    display: block;
}

.dropdown-submenu>a:after {
    display: block;
    content: " ";
    float: right;
    width: 0;
    height: 0;
    border-color: transparent;
    border-style: solid;
    border-width: 5px 0 5px 5px;
    border-left-color: #ccc;
    margin-top: 5px;
    margin-right: -10px;
}

.dropdown-submenu:hover>a:after {
    border-left-color: #fff;
}

.dropdown-submenu.pull-left { /* csslint allow: adjoining-classes */
    float: none;
}

.dropdown-submenu.pull-left>.dropdown-menu { /* csslint allow: adjoining-classes */
    left: -100%;
    margin-left: 10px;
    -webkit-border-radius: 6px 0 6px 6px;
    -moz-border-radius: 6px 0 6px 6px;
    border-radius: 6px 0 6px 6px;
}